#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/time.h>
#ifndef USE_NETLIB
#include <net/ethernet.h>
#include <netinet/in.h>
#ifdef __linux__
#include <netinet/ether.h>
#endif
#include <arpa/inet.h>
#else
#include <netlib.h>
#endif

#include "pktlib.h"
#include "pktbuf.h"

struct macaddr_entry {
  struct ether_addr macaddr;
  pktif_t pktif;
};

#define MACADDR_TABLE_ALLNUM  (1024*4)
#define MACADDR_TABLE_LINENUM (16*4)
#define MACADDR_TABLE_COLNUM  (MACADDR_TABLE_ALLNUM / MACADDR_TABLE_LINENUM)

#define MACADDR_TABLE_HASH(macaddr) \
	( \
		((unsigned int)((unsigned char *)(macaddr))[2] << 24) | \
		((unsigned int)((unsigned char *)(macaddr))[3] << 16) | \
		((unsigned int)((unsigned char *)(macaddr))[4] <<  8) | \
		((unsigned int)((unsigned char *)(macaddr))[5] <<  0) \
	)
#define MACADDR_TABLE_INDEX(macaddr) \
	(MACADDR_TABLE_HASH(macaddr) % MACADDR_TABLE_LINENUM)

static struct macaddr_entry macaddr_table[MACADDR_TABLE_LINENUM][MACADDR_TABLE_COLNUM];
static unsigned int macaddr_hash[MACADDR_TABLE_LINENUM][MACADDR_TABLE_COLNUM];

int macaddr_table_init()
{
  memset(macaddr_table, 0, sizeof(macaddr_table));
  memset(macaddr_hash, 0, sizeof(macaddr_hash));
  return 0;
}

int macaddr_entry_dump()
{
  struct macaddr_entry *entry;
  int index, i;

  printf("---- MAC Address Table ----\n");

  for (index = 0; index < MACADDR_TABLE_LINENUM; index++) {
    for (i = 0; i < MACADDR_TABLE_COLNUM; i++) {
      entry = &macaddr_table[index][i];
      if (entry->pktif) {
	printf("LINE[%04d]:\tMAC Address: %s (Interface: %s)\n", index,
	       ether_ntoa(&entry->macaddr), pktif_get_name(entry->pktif));
      }
    }
  }
  fflush(stdout);

  return 0;
}

struct macaddr_entry *macaddr_entry_search(char macaddr[])
{
  struct macaddr_entry *entry;
  unsigned int hash;
  int index, i;

  hash = MACADDR_TABLE_HASH(macaddr);
  index = MACADDR_TABLE_INDEX(macaddr);

  for (i = 0; i < MACADDR_TABLE_COLNUM; i++) {
    if (macaddr_hash[index][i] == hash) {
      entry = &macaddr_table[index][i];
      if (!memcmp(macaddr, &entry->macaddr, ETHER_ADDR_LEN))
	return entry;
    }
  }

  return NULL;
}

struct macaddr_entry *macaddr_entry_create(char macaddr[])
{
  struct macaddr_entry *entry;
  int index, i;

  index = MACADDR_TABLE_INDEX(macaddr);

  for (i = 0; i < MACADDR_TABLE_COLNUM; i++) {
    entry = &macaddr_table[index][i];
    if (entry->pktif == NULL) {
      memcpy(&entry->macaddr, macaddr, ETHER_ADDR_LEN);
      macaddr_hash[index][i] = MACADDR_TABLE_HASH(macaddr);
      return entry;
    }
  }

  return NULL;
}

struct macaddr_entry *macaddr_entry_add(char macaddr[], pktif_t pktif)
{
  struct macaddr_entry *entry;

  entry = macaddr_entry_search(macaddr);
  if (entry == NULL)
    entry = macaddr_entry_create(macaddr);
  if (entry && (entry->pktif != pktif)) {
    entry->pktif = pktif;
    macaddr_entry_dump();
  }

  return entry;
}

int main(int argc, char *argv[])
{
  pktif_t pktif, p;
  int r, rate = 0;
  pktbuf_t pktbuf;
  int type;
  struct header_param param;
  struct macaddr_entry *entry;

  argc--; argv++;

  if (isdigit(argv[0][0])) { /* ͤꤵ줿 */
    rate = atoi(argv[0]); /* ѥåѴΨ */
    argc--; argv++;
    srand(time(NULL));
  }

  while (argc > 0) { /* ʣ󥿡ե򥪡ץ󤹤 */
    /* 󥿡ե򥪡ץ󤹤(ȯѥåȤϼʤ) */
    pktif_open(argv[0], PKTIF_OPEN_FLAG_RECV_NOTSENT, 0);
    argc--; argv++;
  }

  macaddr_table_init();

  while (1) {
    pktif = pktlib_iflist_select(-1); /* ʣ󥿡եԤ */
    if (pktif == NULL) {
      continue;
    }

    pktbuf = pktbuf_recv(pktif, 0); /* ѥåȤ */
    if (pktbuf == NULL)
      continue;

    if ((rate > 0) && ((rand() % rate) == 0))
      goto discard; /* ꤵ줿ΨǥѥåȤѴ */

    type = pktif_get_linktype(pktif);
    if (type != DLT_EN10MB)
      goto discard;

    r = pktbuf_analyze_ethernet(pktbuf, &param);
    if (r < 1)
      goto discard;

    /* MACɥ쥹ؽ */
    macaddr_entry_add(param.ethernet.src_macaddr, pktif);

    /* ѥåȤ */
    entry = macaddr_entry_search(param.ethernet.dst_macaddr);
    if ((param.ethernet.dst_macaddr[0] & 1) || (entry == NULL)) {
      /* ֥ɥ㥹ȤݡȤ */
      for (p = pktlib_iflist_get_list(); p; p = pktif_get_next(p)) {
	if (p != pktif)
	  pktbuf_send(p, pktbuf_clone(pktbuf));
      }
    } else {
      pktbuf = pktbuf_send(entry->pktif, pktbuf);
    }

discard:
    pktbuf_destroy(pktbuf);
  }

  return 0;
}
